"""
I2C Sensor Library

YOOVO Python library for managing I2C sensors with dynamic loading capabilities
and resource management.
"""

from .i2c_bus_manager import I2CBusManager
from .ads_wrapper import ADS1115Wrapper

__version__ = "1.0.0"
__all__ = [
    "I2CBusManager",
    "ADS1115Wrapper",
]



from yoovo.shared.bypassed import _YoovoBypassedSensor
from yoovo.app import get_registry
import logging
import traceback

logger = logging.getLogger("yoovo.hardware")


def _safe_create(name, factory, *args, **kwargs):
    registry = get_registry()
    try:
        instance = factory(*args, **kwargs)
        registry.register(name, instance)
        return instance
    except Exception as e:
        instance = _YoovoBypassedSensor()
        registry.register(name, instance, bypassed_reason=str(e))
        error_msg = (
            f"\n{'='*60}\n"
            f"  HARDWARE ERROR: {name} failed to initialize\n"
            f"  Reason: {e}\n"
            f"  Check wiring and connections for {name}.\n"
            f"{'='*60}"
        )
        logger.error(error_msg)
        logger.debug(traceback.format_exc())
        print(error_msg)
        return instance


class LED:
    def __new__(cls, pin, frequency=1000, name=None):
        from yoovo.hardware.rpi.CompiledLedLight import CompiledLedLight
        return _safe_create(name or f"LED_pin{pin}", CompiledLedLight, pin, frequency)


class KnockSensor:
    PUD_UP = "PUD_UP"
    PUD_DOWN = "PUD_DOWN"
    PUD_OFF = "PUD_OFF"

    def __new__(cls, pin, pull=None, name=None):
        from yoovo.hardware.rpi.CompiledKy031Sensor import CompiledKy031Sensor
        pull = pull or CompiledKy031Sensor.PUD_UP
        return _safe_create(name or f"KnockSensor_pin{pin}", CompiledKy031Sensor, pin, pull)


class GasSensor:
    def __new__(cls, digital_pin=None, digital_active_low=True, channel=0, name=None):
        from yoovo.hardware.rpi.CompiledGasSensor import CompiledGasSensor
        return _safe_create(
            name or f"GasSensor_pin{digital_pin}",
            CompiledGasSensor, digital_pin=digital_pin, digital_active_low=digital_active_low, channel=channel
        )


class Button_Sensor:
    PUD_UP = "PUD_UP"
    PUD_DOWN = "PUD_DOWN"
    PUD_OFF = "PUD_OFF"

    def __new__(cls, pin, pull=None, name=None):
        from yoovo.hardware.rpi.CompiledKy004Sensor import CompiledKy004Sensor
        pull = pull or CompiledKy004Sensor.PUD_UP
        return _safe_create(name or f"ButtonSensor_pin{pin}", CompiledKy004Sensor, pin, pull)


class PIR:
    PUD_UP = "PUD_UP"
    PUD_DOWN = "PUD_DOWN"
    PUD_OFF = "PUD_OFF"

    def __new__(cls, pin, pull=None, name=None):
        from yoovo.hardware.rpi.CompiledPirSensor import CompiledPirSensor
        return _safe_create(name or f"PIR_pin{pin}", CompiledPirSensor, pin, pull)


class ServoMotor:
    def __new__(cls, mode="continuous", pin=None, name=None):
        from yoovo.hardware.rpi.CompiledServoMotor import CompiledServoMotor
        return _safe_create(name or f"Servo_pin{pin}", CompiledServoMotor, mode, pin)


class StepperMotor:
    def __new__(cls, dir_pin=None, step_pin=None, name=None):
        from yoovo.hardware.rpi.CompiledStepperMotor import CompiledStepperMotor
        return _safe_create(name or "StepperMotor", CompiledStepperMotor, dir_pin=dir_pin, step_pin=step_pin)


class LCD:
    def __new__(cls, address=0x27, cols=16, rows=2, name=None):
        from yoovo.hardware.rpi.CompiledLCD import CompiledLCD
        return _safe_create(name or "LCD", CompiledLCD, address, rows, cols)


class Buzzer:
    def __new__(cls, pin=None, buzzer_type="ACTIVE", active_high=True, name=None):
        from yoovo.hardware.rpi.CompiledBuzzerSensor import CompiledBuzzerSensor
        return _safe_create(name or f"Buzzer_pin{pin}", CompiledBuzzerSensor, pin=pin, buzzer_type=buzzer_type, active_high=active_high)


class RgbLed:
    def __new__(cls, red_pin=None, green_pin=None, blue_pin=None, name=None):
        from yoovo.hardware.rpi.CompiledRgbLed import CompiledRgbLed
        return _safe_create(name or "RgbLed", CompiledRgbLed, red_pin, green_pin, blue_pin)


class TemperatureSensor:
    def __new__(cls, pin=None, name=None):
        from yoovo.hardware.rpi.CompiledKy001Sensor import CompiledKy001Sensor
        return _safe_create(name or "TempSensor", CompiledKy001Sensor)


class VibrationSensor:
    PUD_UP = "PUD_UP"
    PUD_DOWN = "PUD_DOWN"
    PUD_OFF = "PUD_OFF"

    def __new__(cls, pin, pull=None, name=None):
        from yoovo.hardware.rpi.CompiledKy002Sensor import CompiledKy002Sensor
        return _safe_create(name or f"VibrationSensor_pin{pin}", CompiledKy002Sensor, pin, pull)


class HallSensor:
    PUD_UP = "PUD_UP"
    PUD_DOWN = "PUD_DOWN"
    PUD_OFF = "PUD_OFF"

    def __new__(cls, pin, pull=None, name=None):
        from yoovo.hardware.rpi.CompiledKy003Sensor import CompiledKy003Sensor
        return _safe_create(name or f"HallSensor_pin{pin}", CompiledKy003Sensor, pin, pull)


class LightBarrier:
    PUD_UP = "PUD_UP"
    PUD_DOWN = "PUD_DOWN"
    PUD_OFF = "PUD_OFF"

    def __new__(cls, pin, pull=None, name=None):
        from yoovo.hardware.rpi.CompiledKy010Sensor import CompiledKy010Sensor
        return _safe_create(name or f"LightBarrier_pin{pin}", CompiledKy010Sensor, pin, pull)


class NtcSensor:
    def __new__(cls, channel=0, name=None):
        from yoovo.hardware.rpi.CompiledKy013Sensor import CompiledKy013Sensor
        return _safe_create(name or f"NtcSensor_ch{channel}", CompiledKy013Sensor, channel=channel)


class TempHumiditySensor:
    def __new__(cls, sensor_type="DHT11", pin=None, name=None):
        from yoovo.hardware.rpi.CompiledKy015Sensor import CompiledKy015Sensor
        return _safe_create(name or f"TempHumidity_pin{pin}", CompiledKy015Sensor, sensor_type, pin)


class TiltSwitch:
    PUD_UP = "PUD_UP"
    PUD_DOWN = "PUD_DOWN"
    PUD_OFF = "PUD_OFF"

    def __new__(cls, pin, pull=None, name=None):
        from yoovo.hardware.rpi.CompiledKy017Sensor import CompiledKy017Sensor
        return _safe_create(name or f"TiltSwitch_pin{pin}", CompiledKy017Sensor, pin, pull)


class Photoresistor:
    def __new__(cls, channel=0, name=None):
        from yoovo.hardware.rpi.CompiledKy018Sensor import CompiledKy018Sensor
        return _safe_create(name or f"Photoresistor_ch{channel}", CompiledKy018Sensor, channel)


class Relay:
    def __new__(cls, pin, active_high=True, name=None):
        from yoovo.hardware.rpi.CompiledKy019Sensor import CompiledKy019Sensor
        return _safe_create(name or f"Relay_pin{pin}", CompiledKy019Sensor, pin, active_high)


class ReedSwitch:
    PUD_UP = "PUD_UP"
    PUD_DOWN = "PUD_DOWN"
    PUD_OFF = "PUD_OFF"

    def __new__(cls, pin, pull=None, name=None):
        from yoovo.hardware.rpi.CompiledKy021Sensor import CompiledKy021Sensor
        return _safe_create(name or f"ReedSwitch_pin{pin}", CompiledKy021Sensor, pin, pull)


class Joystick:
    def __new__(cls, x_channel=0, y_channel=1, button_pin=None, name=None):
        from yoovo.hardware.rpi.CompiledKy023Sensor import CompiledKy023Sensor
        return _safe_create(name or "Joystick", CompiledKy023Sensor, x_channel, y_channel, button_pin)


class HallEffectSensor:
    def __new__(cls, digital_pin=None, digital_active_low=True, channel=0, name=None):
        from yoovo.hardware.rpi.CompiledKy024Sensor import CompiledKy024Sensor
        return _safe_create(name or f"HallEffect_pin{digital_pin}", CompiledKy024Sensor, digital_pin=digital_pin, digital_active_low=digital_active_low, channel=channel)


class ReedSensor:
    def __new__(cls, digital_pin=None, digital_active_low=True, channel=0, name=None):
        from yoovo.hardware.rpi.CompiledKy025Sensor import CompiledKy025Sensor
        return _safe_create(name or f"ReedSensor_pin{digital_pin}", CompiledKy025Sensor, digital_pin=digital_pin, digital_active_low=digital_active_low, channel=channel)


class FlameSensor:
    def __new__(cls, digital_pin=None, digital_active_low=True, channel=0, name=None):
        from yoovo.hardware.rpi.CompiledKy026Sensor import CompiledKy026Sensor
        return _safe_create(name or f"FlameSensor_pin{digital_pin}", CompiledKy026Sensor, digital_pin=digital_pin, digital_active_low=digital_active_low, channel=channel)


class ThermistorSensor:
    def __new__(cls, digital_pin=None, digital_active_low=True, channel=0, name=None):
        from yoovo.hardware.rpi.CompiledKy028Sensor import CompiledKy028Sensor
        return _safe_create(name or f"Thermistor_pin{digital_pin}", CompiledKy028Sensor, digital_pin=digital_pin, digital_active_low=digital_active_low, channel=channel)


class TwoColorLed:
    def __new__(cls, red_pin=None, green_pin=None, name=None):
        from yoovo.hardware.rpi.CompiledKy029Sensor import CompiledKy029Sensor
        return _safe_create(name or "TwoColorLed", CompiledKy029Sensor, red_pin, green_pin)


class ObstacleDetector:
    PUD_UP = "PUD_UP"
    PUD_DOWN = "PUD_DOWN"
    PUD_OFF = "PUD_OFF"

    def __new__(cls, pin, pull=None, name=None):
        from yoovo.hardware.rpi.CompiledKy032Sensor import CompiledKy032Sensor
        return _safe_create(name or f"ObstacleDetector_pin{pin}", CompiledKy032Sensor, pin, pull)


class LineTracker:
    PUD_UP = "PUD_UP"
    PUD_DOWN = "PUD_DOWN"
    PUD_OFF = "PUD_OFF"

    def __new__(cls, pin, pull=None, name=None):
        from yoovo.hardware.rpi.CompiledKy033Sensor import CompiledKy033Sensor
        return _safe_create(name or f"LineTracker_pin{pin}", CompiledKy033Sensor, pin, pull)


class FlashLed:
    def __new__(cls, pin, name=None):
        from yoovo.hardware.rpi.CompiledKy034Sensor import CompiledKy034Sensor
        return _safe_create(name or f"FlashLed_pin{pin}", CompiledKy034Sensor, pin)


class MagneticSensor:
    def __new__(cls, channel=0, name=None):
        from yoovo.hardware.rpi.CompiledKy035Sensor import CompiledKy035Sensor
        return _safe_create(name or f"MagneticSensor_ch{channel}", CompiledKy035Sensor, channel=channel)


class TouchSensor:
    PUD_UP = "PUD_UP"
    PUD_DOWN = "PUD_DOWN"
    PUD_OFF = "PUD_OFF"

    def __new__(cls, digital_pin=None, channel=0, pull_up_down=None, name=None):
        from yoovo.hardware.rpi.CompiledKy036Sensor import CompiledKy036Sensor
        return _safe_create(name or f"TouchSensor_pin{digital_pin}", CompiledKy036Sensor, digital_pin=digital_pin, channel=channel, pull_up_down=pull_up_down)


class Microphone:
    def __new__(cls, digital_pin=None, digital_active_low=True, channel=0, name=None):
        from yoovo.hardware.rpi.CompiledKy037Sensor import CompiledKy037Sensor
        return _safe_create(name or f"Microphone_pin{digital_pin}", CompiledKy037Sensor, digital_pin=digital_pin, digital_active_low=digital_active_low, channel=channel)


class HeartbeatSensor:
    def __new__(cls, channel=0, name=None):
        from yoovo.hardware.rpi.CompiledKy039Sensor import CompiledKy039Sensor
        return _safe_create(name or f"Heartbeat_ch{channel}", CompiledKy039Sensor, channel)


class RotaryEncoder:
    PUD_UP = "PUD_UP"
    PUD_DOWN = "PUD_DOWN"
    PUD_OFF = "PUD_OFF"

    def __new__(cls, pin_clk, pin_dt, pin_sw=None, pull=None, name=None):
        from yoovo.hardware.rpi.CompiledKy040Sensor import CompiledKy040Sensor
        return _safe_create(name or "RotaryEncoder", CompiledKy040Sensor, pin_clk, pin_dt, pin_sw, pull)


class UltrasonicSensor:
    def __new__(cls, trigger_pin, echo_pin, name=None):
        from yoovo.hardware.rpi.CompiledKy050Sensor import CompiledKy050Sensor
        return _safe_create(name or "Ultrasonic", CompiledKy050Sensor, trigger_pin, echo_pin)


class MotionSensor:
    def __new__(cls, name=None):
        from yoovo.hardware.rpi.CompiledMpuSensor import CompiledMpuSensor
        return _safe_create(name or "MotionSensor", CompiledMpuSensor)


class SoilMoisture:
    PUD_UP = "PUD_UP"
    PUD_DOWN = "PUD_DOWN"
    PUD_OFF = "PUD_OFF"

    def __new__(cls, pin=None, pull=None, channel=0, name=None):
        from yoovo.hardware.rpi.CompiledSoilMoistureSensor import CompiledSoilMoistureSensor
        return _safe_create(name or f"SoilMoisture_ch{channel}", CompiledSoilMoistureSensor, pin, pull, channel)


class Thermocouple:
    def __new__(cls, pin=None, name=None):
        from yoovo.hardware.rpi.CompiledThermocoupleSensor import CompiledThermocoupleSensor
        return _safe_create(name or "Thermocouple", CompiledThermocoupleSensor, pin)


class WaterLevel:
    def __new__(cls, pin=None, channel=0, name=None):
        from yoovo.hardware.rpi.CompiledWaterLevelSensor import CompiledWaterLevelSensor
        return _safe_create(name or f"WaterLevel_ch{channel}", CompiledWaterLevelSensor, pin, channel)


class LedMatrix:
    def __new__(cls, interface='spi', name=None, **kwargs):
        from yoovo.hardware.rpi.CompiledLedMatrix import CompiledLedMatrix
        return _safe_create(name or "LedMatrix", CompiledLedMatrix, interface=interface, **kwargs)


class PwmOutput:
    def __new__(cls, pin, frequency=1000, name=None):
        from yoovo.hardware.rpi.CompiledPwmOutput import CompiledPwmOutput
        return _safe_create(name or f"PwmOutput_pin{pin}", CompiledPwmOutput, pin, frequency)


class Timer:
    def __new__(cls, interval=1000, is_loop=True, callback=None, name=None):
        from yoovo.hardware.rpi.CompiledTimer import CompiledTimer
        return _safe_create(name or "Timer", CompiledTimer, interval, is_loop, callback)


class AnalogDigitalInput:
    PUD_UP = "PUD_UP"
    PUD_DOWN = "PUD_DOWN"
    PUD_OFF = "PUD_OFF"

    def __new__(cls, pin, pull=None, channel=0, name=None):
        from yoovo.hardware.rpi.CompiledAnalogDigitalInput import CompiledAnalogDigitalInput
        return _safe_create(name or f"AnalogDigitalInput_pin{pin}", CompiledAnalogDigitalInput, pin, pull, channel)


class DigitalInput:
    PUD_UP = "PUD_UP"
    PUD_DOWN = "PUD_DOWN"
    PUD_OFF = "PUD_OFF"

    def __new__(cls, pin, pull=None, name=None):
        from yoovo.hardware.rpi.CompiledDigitalInput import CompiledDigitalInput
        return _safe_create(name or f"DigitalInput_pin{pin}", CompiledDigitalInput, pin, pull)


class AnalogInput:
    def __new__(cls, channel, name=None):
        from yoovo.hardware.rpi.CompiledAnalogInput import CompiledAnalogInput
        return _safe_create(name or f"AnalogInput_ch{channel}", CompiledAnalogInput, channel)



from yoovo.app import app, YoovoApp
from yoovo.screen import Screen
from yoovo.widgets.Button import Button
from yoovo.widgets.Label import Label
from yoovo.widgets.TextBox import TextBox
from yoovo.widgets.Toggle import Toggle
from yoovo.widgets.CheckBox import CheckBox
from yoovo.widgets.ProgressBar import ProgressBar
from yoovo.widgets.Slider import Slider
from yoovo.widgets.Image import Image
from yoovo.widgets.RadioButton import RadioButton
from yoovo.widgets.NumericInput import NumericInput
from yoovo.widgets.ListBox import ListBox
from yoovo.widgets.Chart import Chart
from yoovo.widgets.Panel import Panel
from yoovo.widgets.Video import Video
from yoovo.widgets.ColorPicker import ColorPicker
from yoovo.widgets.DateTimePicker import DateTimePicker
from yoovo.widgets.Loader import Loader
from yoovo.widgets.Layout import Layout

__all__ = [
    'app', 'YoovoApp', 'Screen',
    'Button', 'Label', 'TextBox', 'Toggle', 'CheckBox',
    'ProgressBar', 'Slider', 'Image', 'RadioButton',
    'NumericInput', 'ListBox', 'Chart', 'Panel', 'Video',
    'ColorPicker', 'DateTimePicker', 'Loader', 'Layout',
]



class _YoovoBypassedValue:
    def __call__(self, *args, **kwargs):
        return None

    def __getattr__(self, name):
        return self

    def __bool__(self):
        return False

    def __int__(self):
        return 0

    def __float__(self):
        return 0.0

    def __str__(self):
        return ""


class _YoovoBypassedSensor:
    _value = _YoovoBypassedValue()

    def __getattr__(self, name):
        return self._value

    def __bool__(self):
        return False


__all__ = ["_YoovoBypassedSensor", "_YoovoBypassedValue"]


import inspect
import threading

try:
    import RPi.GPIO as GPIO
except Exception:
    GPIO = None


_LOCK = threading.RLock()
_PATCHED = False
_ORIGINALS = {}
_PIN_STATES = {}
_EVENT_STATES = {}
_PWM_STATES = {}


def _caller_owner_key():
    frame = inspect.currentframe()
    try:
        caller = frame.f_back.f_back
        while caller is not None:
            owner = caller.f_locals.get("self")
            if owner is not None:
                return ("self", id(owner))
            caller = caller.f_back
    finally:
        del frame
    return None


def _callback_owner_key(callback):
    owner = getattr(callback, "__self__", None)
    if owner is not None:
        return ("self", id(owner))
    caller_owner = _caller_owner_key()
    if caller_owner is not None:
        return caller_owner
    return ("callback", id(callback))


def _normalize_pins(channel):
    if channel is None:
        return None
    if isinstance(channel, (list, tuple, set)):
        return [int(pin) for pin in channel]
    return [int(channel)]


def _edge_matches(requested_edge, actual_edge):
    if requested_edge == GPIO.BOTH or actual_edge == GPIO.BOTH:
        return True
    return requested_edge == actual_edge


def _dispatch_pin_event(pin):
    with _LOCK:
        state = _EVENT_STATES.get(pin)
        if not state:
            return
        callbacks = list(state["callbacks"])
        previous_value = state["last_value"]

    try:
        current_value = GPIO.input(pin)
    except Exception:
        current_value = None

    if current_value is None:
        actual_edge = GPIO.BOTH
    elif previous_value is None:
        actual_edge = GPIO.RISING if current_value else GPIO.FALLING
    elif current_value != previous_value:
        actual_edge = GPIO.RISING if current_value else GPIO.FALLING
    else:
        actual_edge = GPIO.BOTH

    with _LOCK:
        state = _EVENT_STATES.get(pin)
        if state is not None:
            state["last_value"] = current_value

    for entry in callbacks:
        if not _edge_matches(entry["edge"], actual_edge):
            continue
        try:
            entry["callback"](pin)
        except Exception:
            pass


def _register_master_event(pin):
    state = _EVENT_STATES.get(pin)
    if not state:
        return

    callbacks = state["callbacks"]
    if not callbacks:
        if state["registered"]:
            try:
                _ORIGINALS["remove_event_detect"](pin)
            except Exception:
                pass
        _EVENT_STATES.pop(pin, None)
        return

    bounce_time = max(int(entry["bouncetime"] or 0) for entry in callbacks)
    if state["registered"] and state["bouncetime"] == bounce_time:
        return

    if state["registered"]:
        try:
            _ORIGINALS["remove_event_detect"](pin)
        except Exception:
            pass

    try:
        _ORIGINALS["add_event_detect"](
            pin,
            GPIO.BOTH,
            callback=lambda channel, _pin=pin: _dispatch_pin_event(_pin),
            bouncetime=bounce_time,
        )
        state["registered"] = True
        state["bouncetime"] = bounce_time
        try:
            state["last_value"] = GPIO.input(pin)
        except Exception:
            state["last_value"] = None
    except Exception:
        state["registered"] = False


def _patched_setup(channel, direction, pull_up_down=None, initial=None):
    pin = int(channel)
    with _LOCK:
        state = _PIN_STATES.get(pin)
        if state is None:
            _ORIGINALS["setup"](pin, direction, pull_up_down=pull_up_down, initial=initial)
            _PIN_STATES[pin] = {
                "refs": 1,
                "direction": direction,
                "pull_up_down": pull_up_down,
                "initial": initial,
            }
            return

        state["refs"] += 1
        if state["direction"] != direction:
            return


def _teardown_pin(pin):
    event_state = _EVENT_STATES.pop(pin, None)
    if event_state and event_state.get("registered"):
        try:
            _ORIGINALS["remove_event_detect"](pin)
        except Exception:
            pass

    pwm_state = _PWM_STATES.pop(pin, None)
    if pwm_state and pwm_state.get("started"):
        try:
            pwm_state["backend"].stop()
        except Exception:
            pass

    try:
        _ORIGINALS["cleanup"](pin)
    except Exception:
        pass


def _patched_cleanup(channel=None):
    pins = _normalize_pins(channel)
    if pins is None:
        with _LOCK:
            pins = list(_PIN_STATES.keys())
            _PIN_STATES.clear()
        for pin in pins:
            _teardown_pin(pin)
        _EVENT_STATES.clear()
        _PWM_STATES.clear()
        return _ORIGINALS["cleanup"]()

    to_cleanup = []
    with _LOCK:
        for pin in pins:
            state = _PIN_STATES.get(pin)
            if state is None:
                to_cleanup.append(pin)
                continue
            state["refs"] -= 1
            if state["refs"] <= 0:
                _PIN_STATES.pop(pin, None)
                to_cleanup.append(pin)

    for pin in to_cleanup:
        _teardown_pin(pin)


def _patched_add_event_detect(channel, edge, callback=None, bouncetime=0):
    if callback is None:
        return _ORIGINALS["add_event_detect"](channel, edge, callback=callback, bouncetime=bouncetime)

    pin = int(channel)
    owner_key = _callback_owner_key(callback)
    with _LOCK:
        state = _EVENT_STATES.setdefault(
            pin,
            {"callbacks": [], "registered": False, "bouncetime": 0, "last_value": None},
        )
        for entry in state["callbacks"]:
            if entry["owner"] == owner_key and entry["callback"] == callback and entry["edge"] == edge:
                entry["bouncetime"] = max(int(entry["bouncetime"] or 0), int(bouncetime or 0))
                break
        else:
            state["callbacks"].append(
                {
                    "owner": owner_key,
                    "callback": callback,
                    "edge": edge,
                    "bouncetime": int(bouncetime or 0),
                }
            )
        _register_master_event(pin)
    return True


def _patched_remove_event_detect(channel):
    pin = int(channel)
    owner_key = _caller_owner_key()
    with _LOCK:
        state = _EVENT_STATES.get(pin)
        if state is None:
            try:
                return _ORIGINALS["remove_event_detect"](pin)
            except Exception:
                return None

        if owner_key is None:
            state["callbacks"] = []
        else:
            state["callbacks"] = [
                entry for entry in state["callbacks"] if entry["owner"] != owner_key
            ]

        if not state["callbacks"]:
            if state["registered"]:
                try:
                    _ORIGINALS["remove_event_detect"](pin)
                except Exception:
                    pass
            _EVENT_STATES.pop(pin, None)
            return None

        _register_master_event(pin)
    return None


class _SharedPWMProxy:
    def __init__(self, pin, state):
        self._pin = pin
        self._state = state
        self._active = True
        self._started = False

    def start(self, duty_cycle):
        with _LOCK:
            if not self._active:
                return
            if not self._state["started"]:
                self._state["backend"].start(duty_cycle)
                self._state["started"] = True
            self._started = True

    def ChangeDutyCycle(self, duty_cycle):
        with _LOCK:
            if not self._active:
                return
            if not self._state["started"]:
                self._state["backend"].start(duty_cycle)
                self._state["started"] = True
            else:
                self._state["backend"].ChangeDutyCycle(duty_cycle)
            self._started = True

    def ChangeFrequency(self, frequency):
        with _LOCK:
            if not self._active:
                return
            self._state["backend"].ChangeFrequency(frequency)

    def stop(self):
        with _LOCK:
            if not self._active:
                return
            self._active = False
            self._state["owners"] -= 1
            if self._state["owners"] <= 0:
                try:
                    self._state["backend"].stop()
                finally:
                    _PWM_STATES.pop(self._pin, None)


def _patched_pwm(channel, frequency):
    pin = int(channel)
    with _LOCK:
        state = _PWM_STATES.get(pin)
        if state is None:
            state = {
                "backend": _ORIGINALS["PWM"](pin, frequency),
                "owners": 0,
                "started": False,
            }
            _PWM_STATES[pin] = state
        state["owners"] += 1
        return _SharedPWMProxy(pin, state)


def ensure_gpio_patched():
    global _PATCHED

    if GPIO is None:
        return False

    with _LOCK:
        if _PATCHED:
            return True

        _ORIGINALS["setup"] = GPIO.setup
        _ORIGINALS["cleanup"] = GPIO.cleanup
        _ORIGINALS["add_event_detect"] = GPIO.add_event_detect
        _ORIGINALS["remove_event_detect"] = GPIO.remove_event_detect
        _ORIGINALS["PWM"] = GPIO.PWM

        GPIO.setup = _patched_setup
        GPIO.cleanup = _patched_cleanup
        GPIO.add_event_detect = _patched_add_event_detect
        GPIO.remove_event_detect = _patched_remove_event_detect
        GPIO.PWM = _patched_pwm
        _PATCHED = True
        return True


__all__ = ["ensure_gpio_patched"]